% PART 1: Initialize the code

clear
scrsz = get(0,'ScreenSize'); % get the current screen resolution
figure(3);
if figure(1)
    clf(1,'reset');
end
set(figure(1),'Name','Inversion of Geoacoustic Parameters using Genetic Algorithms','NumberTitle','on','Position',[40 100 scrsz(3)/2.1 round(scrsz(4)/1.3)],'Tag','Plotting Figure');
start_button = uicontrol('String','Start','Position',[scrsz(3)/2.2-40 scrsz(4)/3 72 36],'Enable','on','callback','[push1]=pushbutton(push1)');
end_button = uicontrol('String','End','Position',[scrsz(3)/2.2-40 scrsz(4)/3-50 72 36],'Enable','on','callback','[push]=pushbutton(push)');
pause_button = uicontrol('String','Pause','Position',[scrsz(3)/2.2-40 scrsz(4)/3-100 72 36],'Style','checkbox','Enable','on','callback','[pausebutton]=pushbutton(pausebutton)');

% initialize; set boundaries, precision, etc:

%
%
drawnow
push1=0;

while push1==0              % Press "Start" to start the inversion.
    pause(0.1);
end

presga      % get the inversion parameters, boundaris and discretizing.

pausebutton=0;           
push=0;

% PART 2: Set inversion options

% Construction of the Field structure:
Field = struct('length',bits,...
                'code',1,...
                'bound',1,...
                'lb',lower_bounds,...
                'ub',upper_bounds);

%				[length;	(vector)
%				 code;		(0=binary     | 1=graycoded binary)
%                bound;     (0=No         | 1=Yes
%				 lb;		(num)
%				 ub;		(num)
%
%			   where
%				length   - row vector containing the length (bitprecision),
%                   of each substring in Chrom. sum(lenth)
%					should equal the individual length.
%				code  - binary row vector indicating how each
%					substring is to be decoded.
%               bound    - Include boundaries
%				lb,
%				ub    - Vector containing Lower and upper bounds for each
%					parameter to be estimated. 

% Construction of the Options structure:         
Options = struct('popsize',56,...
                 'lchrom',sum(Field.length),...
                 'maxgen',60,...
                 'pcross',0.8,...
                 'pmutate',0.05,...
                 'nsubpop',1,...
                 'gengap',0.5);


%Options.pmutate = (1/Options.lchrom);
             
%               popsize      (int)
%                   -  the size of the population, scalar
% 
%               lchrom      (int)
%                   - the length of the chromosones, must equal the sum of
%                     the phenomtypes length, scalar
% 
%               maxgen     (int)
%                   - number of generations too produce, scalar
% 
%               pcross      ([0,1])
%                   - Crossover rate. Scalar between 0 and 1, 
% 
%               pmutate     ([0,1])
%                   - Mutation rate. Scalar between 0 and 1
%               
%               nsubpop     (int)
%                   - Number of subpopulations to be produced, if nsubpop
%                     is NaN, 1 is assumed
% 
%               gengap      ([0, 1])
%                   - fraction of the population to be replaced by
%                     selection
%               


% Make sure the generation gap is a even number
if ceil((Options.popsize*Options.gengap))/2 ~= fix(ceil((Options.popsize*Options.gengap))/2)
    GGAP = ceil(Options.popsize*Options.gengap) -1 ;
else
    GGAP = ceil(Options.popsize*Options.gengap);
end

% Initialize
filedata = readfile('observed.pre');        % Import som info, so we can reproduse .pre files later

nrec = str2double(filedata{2}{3,1}(5:15));  % Number of receivers
nPar = size(Field.length,2);                % number of parameters to estimate

% import the observed frequency spectrum
fid = fopen('observed.out','r');                                        
data = textscan(fid,'%n %n','headerLines', 36, 'delimiter', '\n');
obs_data = complex(data{1,1},data{1,2});

% the size of the generated out files is needed later
temp1=dir('observed.out');
filesize = temp1.bytes;

ncross = 0;
nmutate = 0;
gen = 1;
push=0;
T0=100;

min_phen=[];    
avg=[];
min_gen=[];
BestObj = ones(20,1)*inf;
BestPhen = zeros(20, nPar);

% Delete old files not needed
name = 'model*';
delete(name);




% PART 3:                     Start the inversion!!!

tic;    % Start the timer

%Initialize the first population:
NewChrom = initialize(Options.popsize,Options.lchrom,Options.nsubpop);


PhenOld = decode(NewChrom, Field);              % decode to population to their real values

write_file(PhenOld, filedata, model_matrix, nlayers, nParm, inv_parameters );   %Write pre files for every modelvector


% Run a unix c-shell script to start the forward model OSIRIS, 
% 4 models can be calculated simultanously.
for z = 1:4:Options.popsize
    pause(1)
    fprintf('Processing model %d of %d \n', z, Options.popsize);
    unix(['sgaosiris* model_', num2str(z, '%3.2d'), '.pre &']);
    pause(0.5)
    fprintf('Processing model %d of %d \n', z+1, Options.popsize);
    unix(['sgaosiris* model_', num2str(z+1, '%3.2d'), '.pre &']);
    pause(0.5)
    fprintf('Processing model %d of %d \n', z+2, Options.popsize);    
    unix(['sgaosiris* model_', num2str(z+2, '%3.2d'), '.pre &']);
    pause(0.5)
    fprintf('Processing model %d of %d \n', z+3, Options.popsize);    
    unix(['sgaosiris* model_', num2str(z+3, '%3.2d'), '.pre']);
    
    file1 = dir(['model_', num2str(z, '%3.2d'), '.out']);
    file2 = dir(['model_', num2str(z+1, '%3.2d'), '.out']);
    file3 = dir(['model_', num2str(z+2, '%3.2d'), '.out']);
    tot_size = file1.bytes + file2.bytes + file3.bytes;
    if tot_size < (3*filesize)  % Make sure all the calculating is done before proceeding
        while 1
            fprintf('Wainting for OSIRIS...\n');
            pause(0.5);
            file1 = dir(['model_', num2str(z, '%3.2d'), '.out']);
            file2 = dir(['model_', num2str(z+1, '%3.2d'), '.out']);
            file3 = dir(['model_', num2str(z+2, '%3.2d'), '.out']);
            tot_size = file1.bytes + file2.bytes + file3.bytes;
            if tot_size >= (3*filesize)
                break
            end
        end
    end
end
fprintf('\n\nProcessing...\n\n');


data_inn = importdata(Options.popsize);         % Import all the generated frequncy responses

ObjValOld = objfunc2(obs_data, data_inn);

           
[BestObj, BestPhen] = storebest(PhenOld, BestPhen, ObjValOld, BestObj);     %Make sure to save the 20 best

% Calculate some statistics:
max_gen(gen) = max(ObjValOld);
[min_gen(gen), A] = min(ObjValOld);
avg(gen) = mean(ObjValOld);
min_phen(gen,:) = PhenOld(A,:);

OldChrom = NewChrom;

% Its time for some plotting:
figure(1);
% PLOTTS:    
    % Plot the parameter values of the best individual in each generation
for i = 1:nParm    
    subplot(2,nParm,i);
    idx = 1+nlayers*(i-1):nlayers*i;
    switch (inv_parameters(i,:))
        case ('ti')
            idx = idx(1:end-1);
    end    
    bar(PhenOld(A,idx));title('Current best individual');xlabel(['parameter: ', inv_parameters(i,:)]);
    for u=1:size(idx,2)
        text(u-0.36, min_phen(gen,idx(u))*0.9,['\color{white}' num2str(ceil(min_phen(gen,idx(u))),'%.0f')]);
    end
end
    % Plot the value of the objective function, best and average
    subplot(2,nParm,nParm+1:2*nParm); plot(min_gen,'--w.','LineWidth',2,'MarkerEdgeColor','r','MarkerFaceColor','g','MarkerSize',10);
    hold on; plot(avg,':w.','LineWidth',2,'MarkerEdgeColor','b','MarkerFaceColor','b','MarkerSize',10); hold off;
    axis([0 Options.maxgen 0 inf]);
    title(['Best = ', num2str(min_gen(gen)),'   Mean = ',num2str(avg(gen))],'Units','normalized');
    xlabel('Generation'); ylabel('f(x)');
    
    figure(3);  plotgeneration(PhenOld, nParm, nlayers, inv_parameters);
    
    
    
while gen < Options.maxgen        % Main iterative loop (feedback-loop)
        
    gen=gen+1;                    % increase the generation counter

    % Find the best ind. too be kept and later for reinsertion
    [ans,I] = sort(ObjValOld,1,'descend');
    
% CALCULATE EXPECTET NUMBER OF PARENTS FOR EACH INDIVIDUAL (CLACULATE FITNESS):
     p=prob(ObjValOld, min_gen(gen-1),GGAP, gen);        % Calculate with probability based on temperature
%      p = scale(ObjValOld);                               % Linear based fitness ranking  
%      p = fitsrank(ObjValOld, Options.popsize);           % Exponential based fitness ranking  


% CHOOSE PARENTS BASED ON FITNESS:
     parents = sel_roulette(p, GGAP);                    % Roulette Wheel Selection
%     parents = stochastic_selection(GGAP,p,1);          % Remainder stochastic selection 
%     parents = sus(p, GGAP);                            % Stochastic universal selection
    
% Shuffle the selected parents before crossover
    parents = randsample(parents,length(parents));
% 
% CROSSOVER:
%      [NewChrom,n] = crossover(OldChrom, Options.pcross, parents);                      % Single point         
%      NewChrom = mulpointxover2(OldChrom, Options.pcross, parents,Field.length);        % multiple point crossover
     NewChrom = xovermulti(OldChrom, Options.pcross, parents, 4);                       % multipoint crossover
    
% MUTATION:
    [NewChrom, nmutate] = mutation(NewChrom, Options.pmutate, nmutate);
    
% DECODING & EVALUATION (Again...):    
    PhenNew = decode(NewChrom, Field);
    
    % Update the alreay created pre-files with the new parameters
    o_write(PhenNew, model_matrix, nlayers, nParm, inv_parameters);

% call osiris and calculate
for z = 1:4:GGAP
    fprintf('Processing model %d of %d \n', z, GGAP);   
    unix(['sgaosiris* model_', num2str(z, '%3.2d'), '.pre &']);
    pause(0.5)
    fprintf('Processing model %d of %d \n', z+1, GGAP);
    unix(['sgaosiris* model_', num2str(z+1, '%3.2d'), '.pre &']);
    pause(0.5)
    fprintf('Processing model %d of %d \n', z+2, GGAP);   
    unix(['sgaosiris* model_', num2str(z+2, '%3.2d'), '.pre &']);
    pause(0.5)
    fprintf('Processing model %d of %d \n', z+3, GGAP);    
    unix(['sgaosiris* model_', num2str(z+3, '%3.2d'), '.pre']);

    file1 = dir(['model_', num2str(z, '%3.2d'), '.out']);
    file2 = dir(['model_', num2str(z+1, '%3.2d'), '.out']);
    file3 = dir(['model_', num2str(z+2, '%3.2d'), '.out']);
    tot_size = file1.bytes + file2.bytes + file3.bytes;
    if tot_size < (3*filesize)
        while 1
            fprintf('Wainting for OSIRIS...\n');
            pause(0.5);
            file1 = dir(['model_', num2str(z, '%3.2d'), '.out']);
            file2 = dir(['model_', num2str(z+1, '%3.2d'), '.out']);
            file3 = dir(['model_', num2str(z+2, '%3.2d'), '.out']);
            tot_size = file1.bytes + file2.bytes + file3.bytes;
            if tot_size >= (3*filesize)
                break
            end
        end
    end
end
    fprintf('\nProcessing\n\n');
    
    data_inn_new = importdata(GGAP);
%    data_inn_new = process(data_inn_new, X, nrec);
    ObjValNew = objfunc2(obs_data, data_inn_new); 

        
% REINSERTION:
    OldChrom(I(1:GGAP),:) = NewChrom(:,:);
    ObjValOld(I(1:GGAP),:) = ObjValNew(:,:);
    PhenOld(I(1:GGAP),:) = PhenNew(:,:);
    data_inn(:,I(1:GGAP)) = data_inn_new(:,:);

    [BestObj, BestPhen] = storebest(PhenOld, BestPhen, ObjValOld, BestObj);
    
% STATISTICS:
    [min_gen(gen), A] = min(ObjValOld);
    avg(gen) = mean(ObjValOld);
    min_phen(gen,:) = PhenOld(A,:);

% PLOTTS:    
    % Plot the value of the objective function, best and average
    figure(1);
     for i = 1:nParm    
        subplot(2,nParm,i);
        idx = 1+nlayers*(i-1):nlayers*i;
        switch (inv_parameters(i,:))
            case ('ti')
                idx = idx(1:end-1);
        end
        bar(PhenOld(A,idx));title('Current best individual');xlabel('Layer');
        title(['parameter: ', inv_parameters(i,:)])
        for u=1:size(idx,2)
            text(u-0.36, min_phen(gen,idx(u))*0.9,['\color{white}' num2str(ceil(min_phen(gen,idx(u))),'%.0f')]);
        end
    end
    % Plot the value of the objective function, best and average
    subplot(2,nParm,nParm+1:2*nParm); semilogy(min_gen,'--w.','LineWidth',2,'MarkerEdgeColor','r','MarkerFaceColor','g','MarkerSize',10);
    hold on; semilogy(avg,':w.','LineWidth',2,'MarkerEdgeColor','b','MarkerFaceColor','b','MarkerSize',10); hold off;
    axis([0 Options.maxgen 0 inf]);
    title(['Best = ', num2str(min_gen(gen)),'   Mean = ',num2str(avg(gen))],'Units','normalized');
    xlabel('Generation'); ylabel('f(x)');    
    
     t=toc;
    tb=annotation('textbox',[0.6 0.3 0.15 0.1],'BackgroundColor','w','String',...
    ['Cpu-time: ', num2str(floor(t/60)),' m ',num2str(round(rem(t,60)))]);
    % Plot alle modeller i hver generajson og grupper etter parameter
    figure(3);  plotgeneration(PhenOld, nParm, nlayers, inv_parameters);
    %bar(PhenOld','group'); set(gca,'YTick',[0:100:3200]); grid on;
    
%     set(axes,'YGrid','on');
    grid on;
    drawnow;
    
    if push
        gentemp=gen;
        gen=Options.maxgen;
    elseif pausebutton 
        while pausebutton
            pause(0.1);
        end
    end
end
if push
    gen=gentemp;
end


pause(0.1);

fclose('all');

filename = savevar
save(filename)
pause(0.1);

clear;
